#!/usr/bin/perl -w
# gps-maintain.pl Database maintenance script for gps
#
# Copyright (C) 2004-2005 Michael Moritz <mimo/at/restoel.net>
#
# Contributors: Xin Li <delphij/at/msn.com>
#               Michael Hubbard <mkhubbard/at/gmail.com>
#
# 0.5 2005-02-14 implemented improved 'db_type' => DBI driver name mapping
#                made dsn connection string more generic, should work with all
#                perl DBI database drivers.
#                added 'db_port' option. not currently supported by gps itself.
#                added 'default' functionality to display number of matches.
#                moved display of usage error message in an attempt to make
#                them more noticable.
#
# 0.4 2005-02-14 complete rewrite by Michael Hubbard
# 0.3 2004-09-14 included patch by Xin LI <delphij/at/msn.com>:
#                replaced hardcoded dbname 'greylist' with
#                value from config
# 0.2 2004-07-06 added age only functionality [mimo]

use strict;
use Getopt::Long qw(:config permute);
use DBI;

my @args        = ();
my $configFile  = '';           # i assume this will be defaulted at
                               # some point?
my %gpsConfig   = ();

my %dbiDriverMap = ( 'mysql'   => 'mysql',
                     'pgsql'   => 'Pg',
                     'sqlite'  => 'SQLite',
                     'sqlite2' => 'SQLite2'
                   );

my %dbDefaultPorts = ( 'mysql'   => 3306,
                       'pgsql'   => 5432,
                       'sqlite'  => '',
                       'sqlite2' => ''
                     );

$SIG{__WARN__} = sub {
      if ( $_[0] =~ /^Unknown option:|^Option/ ) {
              # getopt triggered an unknown option
              my $message = lc($_[0]); chomp($message);
              usage("error: $message" );
              return;
      } else {
              # handle warning as usual
              warn @_;
      }
};

#
# sub usage( string $message )
#
# display usage help and optionally display $message
#
sub usage {
      my $message = shift;
      my $bin = `basename $0`; chomp($bin);

      print STDOUT "usage: $bin [options] -age=seconds <config file>\n";
      print STDOUT "\n";
      print STDOUT "options:\n";
      print STDOUT "  -age seconds\t Match triplets older than <seconds>. [required]\n";
      print STDOUT "  -lt|-eq count\t Match triplets with a pass count less than or exactly <count>.\n";
      print STDOUT "  -delete\t Remove matched records from database.\n";
      print STDOUT "  -v,verbose\t Run verbosely.\n";
      print STDOUT "  -debug \t More verbose.\n";
      print STDOUT "\n";
      print STDOUT "\n";
      print STDOUT "See http://mimo.gn.apc.org/gps/index.shtml#dbmaint for more help.\n\n";
      if ( $message ) {
              print STDOUT "*** $message\n";
      }

      exit 1;
}

#
# sub parseGpsConfigFile( string $filename )
#
# open and parse gps configuration file
# -- logic mainly untouched from original version
#
sub parseGpsConfigFile {
      my $filename = shift;
      my %gpsConfig = ();

      open(CONFIG, "<$filename" ) or
              die "can't open $filename: $!";

      while ( <CONFIG> ) {
          chomp;                  # no newline
          s/#.*//;                # no comments
          s/^\s+//;               # no leading white
          s/\s+$//;               # no trailing white
          next unless length;     # anything left?
          my ($var, $value) = split(/\s*=\s*/, $_, 2);

          # remap driver name
          if ( (lc($var) eq 'dbtype') ) {
               my $requestedDriver = lc($value);
               if ( exists($dbiDriverMap{$requestedDriver}) ) {
                       $value = $dbiDriverMap{$requestedDriver};
               } else {
                       die('error: requested use of unsupported DBI driver: '.$requestedDriver);
               }
          }

          $gpsConfig{$var} = $value;
      }

      close(CONFIG);

      # default selected database tcp port if not specified
      if ( !exists($gpsConfig{'db_port'}) ) {
               $gpsConfig{'db_port'} = $dbDefaultPorts{$gpsConfig{'dbtype'}};
      }

      # fix up the sqlite path/dbname stuff
      if ( $gpsConfig{'dbtype'} eq 'SQLite' ) {
              if ( !exists($gpsConfig{'db_sqlite_dbdir'}) ) {
                      # i don't like using die, it's output can be
                      # ugly and terse for non-programmers, but oh
                      # well.
                      die("error: database driver 'sqlite' specified but required 'db_sqlite_dbdir' option not specified in $filename.\n");
              }
              $gpsConfig{'db_dbname'} = $gpsConfig{'db_sqlite_dbdir'}.'/'.$gpsConfig{'db_dbname'};
      }

      return %gpsConfig;
}

#
# sub saveArgument( mixed $arg )
#
# store arguments passed from the command line which do not
# appear to be options.
#
sub saveArgument {
      $args[++$#args] = shift;
}

#
# END SUBS
#

if ( $#ARGV == -1 ) {
      # no command line options, show usage
      usage();
}

# this hash will hold the results from Getopt::Long::GetOptions
# initialize any non-required option variables here.
#
# you were warned.
#
my %opts = ( 'verbose' => 0, 'debug' => 0 );

# parse command line options
#
# note: ordering here is important. verbose|v will toggle the value
# of $opts{'verbose'} for -verbose and -v, were you to swap them to
# v|verbose the value $opts{'v'} would be set and all my tests against
# $opts{'verbose'} will fail, this goes for delete as well.
#
GetOptions( \%opts,
          'verbose|v',
          'debug',
          'delete|d',
          'eq=i',
          'lt=i',
          'age=i',
          'h|?|help' => sub { usage(); },  
          '<>' => \&saveArgument
        );

if ( $#args < 0 ) {
      usage('error: must specify the location of your configuration file. ');
}

# pop first extra arg off of @args and assume it's
# our configuration file.
$configFile = shift(@args);

# verify that required opt --age was provided
if ( !exists($opts{'age'}) ) {
      usage('error: must specify required option value \'-age\'');
}

# verify that we only have '-lt' or '-eq' option
if ( exists($opts{'lt'}) && exists($opts{'eq'}) ) {
      usage('error: please only specify \'-lt\' OR \'-eq\'');
}

# load and parse gps configuration file
%gpsConfig = parseGpsConfigFile( $configFile );

if ( $opts{'debug'} ) {
      # display results of getopt pass
      print STDERR ">>> BEGIN DEBUG DUMP <<<\n\n";
      print STDERR "Getopt::Long Results\n";
      print STDERR "--\n";
      foreach my $key ( keys %opts ) {
              print STDERR "$key\t= $opts{$key}\n";
      }
      print STDERR "\nExtra arguments:\n";
      for( my $i=0;$i<=$#args; $i++) {
              print STDERR "  $i: $args[$i]\n";
      }

      # display results of parseGpsConfigFile()
      print STDERR "\n";
      print STDERR "Configuration values from: $configFile\n";
      foreach my $key ( keys %gpsConfig ) {
              print STDERR "$key   \t= $gpsConfig{$key}\n";
      }
      print STDERR "\n>>> END DEBUG DUMP <<<\n\n";
}

# build sql query criteria
my $sql_clause = "";
my $uts = time() - $opts{'age'};

if ( exists($opts{'lt'}) ) {
      # ??
      $sql_clause = " WHERE uts < $uts AND count < $opts{'lt'}";
} elsif ( exists($opts{'eq'}) ) {
      # ??
      $sql_clause = " WHERE uts < $uts AND count = $opts{'eq'}"; 
} else {
      $sql_clause = " WHERE uts < $uts";
}

if ( $opts{'verbose'} ) { print STDOUT "sql clause: $sql_clause\n"; }

# build dsn connection string
if ( !exists($gpsConfig{'db_host'}) ) {
       # assume localhost
       $gpsConfig{'db_host'} = '127.0.0.1';
}

my $dsn = sprintf('DBI:%s:dbname=%s;host=%s;port=',
                       $gpsConfig{'dbtype'},
                       $gpsConfig{'db_dbname'},
                       $gpsConfig{'db_host'},
                       $gpsConfig{'db_port'});

my $user = $gpsConfig{'db_username'};
my $pass = $gpsConfig{'db_password'};

if ( $opts{'verbose'} ) {
      print STDOUT "using dsn: $dsn\n";
      print STDOUT "username : $user\n";
      print STDOUT "password : <not shown>\n";
}

# connect to database
my $dbh = DBI->connect($dsn, $user, $pass ) or die('error: unable to establish a connection with the database.\n' . DBI->errstr);

if ( $opts{'verbose'} ) { print STDOUT "db: connection established\n"; }

# here we go ...
if ( $opts{'verbose'} ) {
      # display entries before (if) they are removed
      my $queryStr = "SELECT * FROM triplet $sql_clause";
      print STDOUT "$gpsConfig{'dbtype'}: $queryStr\n";

      my $sth = $dbh->prepare($queryStr) or
              die("$gpsConfig{'dbtype'}: unable to prepare statement: ".$dbh->errstr);

      $sth->execute() or
              die("$gpsConfig{'dbtpye'}: unable to execute statement: ".$sth->errstr);

      my $count = 1;
      while ( my @data = $sth->fetchrow_array() ) {
              print STDOUT "$count: ";
              my $i = 0;
              while($i < 5) {
                      print STDOUT $data[$i].",  ";
                      $i++;
              }
              $count++;
              print STDOUT "\n";
      }

      $sth->finish;
}

# and i'll see you on the darkside of the moon ...
if ( $opts{'delete'} ) {

      my $queryStr = "DELETE FROM triplet $sql_clause";

      my $affectedRows = $dbh->do( $queryStr ) or
              die("$gpsConfig{'dbtype'}: error removing records: ".$dbh->errstr);

      print STDOUT "Successfully removed $affectedRows row(s).\n";

} elsif ( !$opts{'verbose'} ) {

       # default behavior added for the sanity of users, doing nothing
       # seems, well, devious.

       my $queryStr = "SELECT * FROM triplet $sql_clause";

       my $sth = $dbh->prepare($queryStr) or
              die("$gpsConfig{'dbtype'}: unable to prepare statement: ".$dbh->errstr);

       $sth->execute() or
              die("$gpsConfig{'dbtpye'}: unable to execute statement: ".$sth->errstr);

       # for some reason $sth->rows returns -1 with sqlite, so we are
       # going to just count them. since this 'mode' will not be ran
       # from a cron job i'm not worried about performance here.

       my $rows = 0;
       while ( my @junk = $sth->fetchrow_array() ) {
               ++$rows;
       }

       print STDOUT "$rows row(s) matched the specified criteria.\n\n".
                    "Use the '-delete' or '-d' option to permenatly remove ".
                    "these records.\nIf you wish to review these records use ".
                    "the '-verbose' option.\n\n";

       $sth->finish;

}

$dbh->disconnect();

if ( $opts{'verbose'} ) { print STDOUT "db: disconnected.\n"; }

exit 0;
